/* Author: Domagoj Babic 2004.
 *
 * $Id: datas.h,v 1.11 2005/03/29 05:29:02 root Exp $
 *
 */

/***********************************************************************
*
*  Copyright (C) 2004-2005, by Domagoj Babic (Author). All rights 
*  reserved. By using this software the USER indicates that he or she 
*  has read, understood and will comply with the following:
*
*  This software may not be called "HyperSAT" if it has been modified in 
*  any way, without the specific prior written permission of the Author.
*
*  Author hereby grants USER nonexclusive permission to use, copy and/or 
*  modify this software for internal, noncommercial, research purposes 
*  only. Any distribution, including commercial sale or license, of this 
*  software, copies of the software, its associated documentation and/or 
*  modifications of either is strictly prohibited without the prior 
*  consent from the Author. Appropriate copyright notice shall be placed 
*  on all software copies, and a complete copy of this notice shall be 
*  included in all copies of the associated documentation.
*
*  THE SOFTWARE IS PROVIDED "AS-IS" AND AUTHOR MAKES NO REPRESENTATIONS 
*  OR WARRANTIES, EXPRESS OR IMPLIED, BY WAY OF EXAMPLE, BUT NOT 
*  LIMITATION. AUTHOR MAKES NO REPRESENTATIONS OR WARRANTIES OF 
*  MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE
*  USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS
*  TRADEMARKS OR OTHER RIGHTS. AUTHOR SHALL NOT BE LIABLE FOR ANY
*  LIABILITY OR DAMAGES WITH RESPECT TO ANY CLAIM BY LICENSEE OR ANY
*  THIRD PARTY ON ACCOUNT OF, OR ARISING FROM THE LICENSE, OR ANY
*  SUBLICENSE OR USE OF THE SOFTWARE OR ANY SERVICE OR SUPPORT.
*
*  Author shall not be liable under any circumstances for any direct, 
*  indirect, special, incidental, or consequential damages with respect 
*  to any claim by USER or any third party on account of or arising from 
*  the use, or inability to use, this software or its associated 
*  documentation, even if the Author has been advised of the possibility 
*  of those damages.
*
*  The above copyright notice must appear in all copies of the software.
*
************************************************************************
*
*  Support notice: Unfortunatelly, the author can't guarantee any 
*  support. The solver has been written exclusively for research 
*  and experimental purposes, so documentation and comments are scarce.
*  Most likely, this will not change in the near future, but feel free
*  to check my webpage occasionally, as new releases (if there will be
*  any) will be made available at my website:
*  http:/www.cs.ubc.ca/~babic/
*
***********************************************************************/

#ifndef H_DATAS
#define H_DATAS

#include <limits.h>
#include <time.h>
#include <math.h>
#include "utils.h"
#include "sort.h"
#include "queues.h"

// ************* CLAUSE DATA STRUCTURE *******************************

// first six bits are reserved for flags
enum clflags_e {EQV_SIGN = 1, LEARNED = 2, EQUIVALENCY = 4,
    TAUTOLOGY = 8, RESERVED1 = 16, RESERVED2 = 32};

enum model_outp_format_e {VECT, READ, RAW};

enum priority_types_e {PRI_APPROX, PRI_EXACT, PRI_SIMPLIFIED,
    PRI_BIN_CLS};

enum scheduled_typ_e {SCHD_DECISION = 0, SCHD_HC_STEM = 1, SCHD_HC_BRANCH = 2};

typedef uint cla_t; 

// max number of literals in a clause : 
// 32 bit machines = 2^14 - 1
// 64 bit machines = 2^30 - 1
#define MAX_LINE_LENGTH 4096 
#define MAX_VARIABLES (UINT_MAX >> 1)         // 2147483647
#define MAX_CL_ENC_BITS (sizeof(uint)*4)    // 16
#define MAX_CL_SIZE ((uint)((1 << MAX_CL_ENC_BITS) - 1))  // 65535
#define CL_FIELD_MASK MAX_CL_SIZE
#define CL_FLAG_MASK 127
#define CL_FLAG_BITS 6
#define CL_USAGE_INCREMENT (1 << CL_FLAG_BITS)

#define UTABLE_FIELD  uintsh //uint_fast8_t

typedef struct table_t {
    unsigned size;
    UTABLE_FIELD cnt;
    unsigned char cell_size;
    unsigned char mask;
    UTABLE_FIELD a[];
} table_t;

#define FLEX_SHRINKED_SIZE 8

typedef struct flexarray_t {
    uint size;
    uint idx;
    uint *a;
} flexarray_t;

enum bcpq_e {BCP_OK, BCP_EXISTS, BCP_CONFLICT};

typedef struct bcpq_t {
    table_t *t;
    uint ridx;
    uint widx;
    cla_t **reasons;
    uint a[];
} bcpq_t;

typedef struct assign_t {
    uint current_dlev;
    uint *todo;
    uint *dec;
    uint tptr;
    uint trd_ptr;
    uint size;
    // assignment queue
    uint didx;
    uint aidx;
    uint a[];
} assign_t;

typedef struct stats_t {
    // preprocessing
    long ppBinEqs;
    long ppTautologies;
    long ppNumEqs;
    long ppUnitVars;
    long ppPureLit;
    long ppNumBinCls;
    long ppTotVarElim;
    long ppTotClsElim;
    uint ppNumClusters;
    double ppTimeCluster;
    double ppTimeGraphGen;
    // clause stats
    long clNumLearned;
    uint clOrigClauses;
    // decision stats
    uint decMaxLev;
    long decNum;
    uint decNumFixedLearn;
    uint decMaxAssigned;
    // s/h-cube stats
    long hcTotBranches;
    long hcTotNodes;
    long hcUseBranches;
    long hcUseNodes;
    long hcNumHcubes;
    long hcIntersections;
    long hcIntersectConflicts;
    float hcAvgBranches;
    float hcAvgNodes;
    float hcAvgUsedBranches;
    float hcAvgUsedNodes;
    float hcFlipBranchDensity;
    float hcFlipNodeDensity;
    // implication stats
    uint impLongestChain;
    long impNumChains; // number of implication chains
    long impNumImps; // total num of implications
    long impBlocksScheduled;
    float impAvgSchBlockSz;
    long impTotalBlockSz;
    float impAvgChainLen; // average chain length
    // conflicts
    long cnfNum;
    long cnfUIPs;
    long cnfUIPCumulativeSz;
    float cnfUIPAvgSz;
    long cnfDECs;
    long cnfDECCumulativeSz;
    float cnfDECAvgSz;
    // misc
    uint varNum;
    uint cacheSize;
    uint cacheDelPerc;
    uint restarts;
    float cacheInflationRate;
    /* resolutions
    long resTotalNum;
    uint resAvgNum;
    // */
} stats_t;

// *************** DECLARATIONS *********************************

inline uint isEquiv(cla_t *cl);
inline uint makeLit(const uint v, uint sgn);
inline uint formVar(const int lit);
inline uint isLearned(cla_t *cl);
inline void markLearned(cla_t *cl);
inline void markTautology(cla_t *c);
inline uint isTautology(cla_t *c);
inline void markEquiv(cla_t *cl);
inline void clearEquiv(cla_t *cl);
inline void setEquivSgn(cla_t *cl, uint sgn);
inline uint getEquivSgn(cla_t *cl);
inline uint getClSize(const cla_t *cl);
inline void setClSize(cla_t *cl, const uint sz);
inline int findVar(const cla_t *cl, const uint lit);
inline int findLit(const cla_t *cl, const uint lit);
inline void delLitAt(cla_t *cl, uint idx);
inline void delLit(cla_t *cl, const uint lit);
inline void replaceLit(cla_t *cl, const uint l1, const uint l2);
inline void clDif(cla_t *a, cla_t *b, uint *dif, uint *nlt, uint *ins);
inline uint isUnitCl(const cla_t *cl);
cla_t *formEqvCla(uint *a, uint sgn);
cla_t *getNewClause(uint *a, uint size);
cla_t *getNewEqvCla(cla_t *c, uint sgn);
#ifdef DEBUG
char checkClSort(cla_t *c);
#endif
inline uint getVar(const uint lit);
inline uint getPhase(const uint lit);
void arrayClPrn(cla_t *cl);
table_t *initTable(uint elements, uint cell_size);
void freeTable(table_t *t);
inline void incrTimeStamp(table_t *t);
inline uint readTable(const table_t *t, const uint idx);
inline uint writeTable(table_t *t, const uint idx, const uint value);
inline uint getFlexArraySize(const flexarray_t *f);
inline uint *getFlexArray(const flexarray_t *f);
#ifdef DEBUG
char checkSortFlexArray(const flexarray_t *f);
#endif
inline void resetFlexArray(flexarray_t *f);
//inline void shrinkFlexArray(flexarray_t *f, uint tresh);
void freeFlexArray(flexarray_t *f);
void delFlexElem(flexarray_t *f, uint i);
void appendToFlexArray(flexarray_t *f, uint x);
flexarray_t *createFlexArray(uint *a, uint sz);
void qsortFlexArray(flexarray_t *f);
inline int searchFlexArray(flexarray_t *f, uint x);
int bfsearchFlexArray(flexarray_t *f, uint x, uint inc);
inline flexarray_t *initFlexArray(uint size);
//inline void swapFlexArray(flexarray_t *a, flexarray_t *b);
//void intersectFlexArrays(flexarray_t *fa, flexarray_t *fb, 
//        flexarray_t *res);
bcpq_t *initBcpq(uint size);
inline int bcpqEmpty(bcpq_t *m);
void freeBcpq(bcpq_t *m);
void bcpqReset(bcpq_t *m);
assign_t *initAQ(uint size);
inline void appendAQ(assign_t *aq, uint lit, uint dlev, char dec);
inline void scheduleAQ(assign_t *aq, uint lit, uint ext);
inline void resetScheduledAQ(assign_t *aq);
inline uint getScheduledAQ(assign_t *aq, uint *lit, uint *ext);
void freeAQ(assign_t *aq);
inline uint numAssignedAQ(assign_t *aq);
inline uint numScheduledAQ(assign_t *aq);
inline uint declev(assign_t *aq);
uint popAQ(assign_t *aq);
inline uint getDecVar(assign_t *aq, uint dlev);
#ifdef COLLECT_STATS
stats_t *computeStats(stats_t *s);
#endif

int hfunc1(uint a, uint b) {
    return (abs((int)a-b) << 3) + (a+1)*(b+1);
}

// init hfunc
int hfunc2(uint a, uint b, uint c) {
    return ((int)(a+1)*(b+1) + abs((int)a-b)*((c << 1)+1)) >> 2;
}

// the following function works great for benchmarks with
// a small number of equivalences
int hfunc3(uint a, uint b) {
    return (abs((int)a-b) << 1) + ((a+b) >> 1);
}

int hfunc4(uint a, uint b) {
    return 8*abs((int)a-b)+((a+b)>>1);
}

int hfunc5(uint a, uint b) {
    return (int)(a+1)*(b+1);
}

int hfunc6(uint a, uint b) {
    return pseudorand((int)(a+1)*(b+1) << 3);
}

//#define init_hfunc(a,b,c) (hfunc(a,b)*(c+1))
//#define init_hfunc(a,b,c) (hfunc(a,b) >> 4)
//#define hfunc(a,b) abs(a-b)+((a+b)>>1)
//#define hfunc(a,b) abs(a-b)*(a+1)*(b+1)
//#define hfunc(a,b) abs(a-b) // 3)
//#define hfunc(a,b) abs(a-b)+(a+b)
//#define hfunc(a,b) pseudorand(10)*abs(a-b)+pseudorand(3)*(a+b) //4)


inline uint isEquiv(cla_t *cl) {
    assert(cl != NULL);
    return cl[0] & EQUIVALENCY;
}

inline uint makeLit(const uint v, uint sgn) {
    assert(v);
    return (v << 1) | sgn;
}

inline uint formVar(const int lit) {
    assert(lit);
    return (lit < 0) ? 
        (uint)abs(lit) << 1 : (uint)((lit) << 1) | 1;
}

inline uint isLearned(cla_t *cl) {
    assert(cl != NULL);
    return cl[0] & LEARNED;
}

inline void markLearned(cla_t *cl) {
    cl[0] |= LEARNED;
}

inline void markTautology(cla_t *c) {
    assert(c != NULL);
    c[0] |= TAUTOLOGY;
}

inline uint isTautology(cla_t *c) {
    assert(c != NULL);
    return c[0] & TAUTOLOGY;
}

inline void markEquiv(cla_t *cl) {
    assert(cl != NULL);
    assert(!isEquiv(cl));
    cl[0] |= EQUIVALENCY;
}

inline void clearEquiv(cla_t *cl) {
    assert(cl != NULL);
    assert(isEquiv(cl));
    cl[0] &= ~EQUIVALENCY;
}

// sgn = 1 -> positive equivalence
// sgn = 0 -> negative
inline void setEquivSgn(cla_t *cl, uint sgn) {
    assert(cl != NULL);
    assert(sgn < 2);
#ifdef UGLY_HACKS
    cl[0] ^= (-sgn ^ cl[0]) & EQV_SIGN;
#ifdef DEBUG
    {
        uint tmp = cl[0];
        if (sgn) {
            assert(cl[0] == (tmp | EQV_SIGN));
        } else {
            assert(cl[0] == (tmp & ~EQV_SIGN));
        }
    }
#endif
#else
    if (sgn) {
        cl[0] |= EQV_SIGN;
    } else {
        cl[0] &= ~EQV_SIGN;
    }
#endif
}

inline uint getEquivSgn(cla_t *cl) {
    assert(cl != NULL);
    return cl[0] & EQV_SIGN;
}

inline uint getUsage(const cla_t *c) {
    return (uint)(c[0] & (CL_FIELD_MASK & ~CL_FLAG_MASK));
}

inline void incUsage(cla_t *c) {
    uint u = (c[0] & CL_FIELD_MASK) + CL_USAGE_INCREMENT;
    c[0] += (u < CL_FIELD_MASK) << CL_FLAG_BITS;
}

inline void resetUsage(cla_t *c) {
    c[0] &= (~CL_FIELD_MASK | CL_FLAG_MASK);
}

#define USG_SCALE 2
inline void scaleUsage(cla_t *c) {
    uint u = (getUsage(c) >> USG_SCALE) & ~CL_FLAG_MASK;
    resetUsage(c);
    c[0] |= u;
}

inline int cl_usgcmp(const void *ca, const void *cb) {
    cla_t *a = (cla_t *)(*((uint *)ca));
    cla_t *b = (cla_t *)(*((uint *)cb));
    // less used clauses first
    return (int)getUsage(a) - getUsage(b);
}

inline int cl_lencmp(const void *ca, const void *cb) {
    cla_t *a = (cla_t *)(*((uint *)ca));
    cla_t *b = (cla_t *)(*((uint *)cb));
    // longer clauses first
    return (int)getClSize(b) - getClSize(a);
}

/*
inline int cl_rlencmp(const void *ca, const void *cb) {
    cla_t *a = (cla_t *)(*((uint *)ca));
    cla_t *b = (cla_t *)(*((uint *)cb));
    uint sa = getClSize(a);
    uint sb = getClSize(b);
    
    if (sa != sb) {
        if (sa == 0) return 1;
        if (sb == 0) return -1;
    }
    
    // shorter clauses first
    return (int)sa - sb;
}// */

// length & usage compare
inline int cl_lucmp(const uint cmp_thold, const void *ca, const void *cb) {
    int len_dif = cl_lencmp(ca, cb);
    return ((uint)abs(len_dif) > cmp_thold) ? len_dif : cl_usgcmp(ca, cb);
}

inline uint getClSize(const cla_t *cl) {
    uintsh *x;
    assert(cl != NULL);
    x = (uintsh *)cl;
    return (uint)x[1];
}

inline void setClSize(cla_t *cl, const uint sz) {
    assert(cl != NULL);
    assert(sz <= MAX_CL_SIZE);
    cl[0] = (cl[0] & CL_FLAG_MASK) | (sz << MAX_CL_ENC_BITS);
}

inline int findVar(const cla_t *cl, const uint lit) {
    return exorsearch((int *)&(cl[1]), (int)(getClSize(cl) - 1), (int)lit) + 1;
}

inline int findLit(const cla_t *cl, const uint lit) {
    return binsearch((int *)&(cl[1]), (int)(getClSize(cl) - 1), (int)lit) + 1;
}

inline void delLitAt(cla_t *cl, uint idx) {
	uint sz;
    assert(idx < getClSize(cl));
    sz = getClSize(cl) - 1;
    if (idx < sz) {
        memmove(&(cl[idx]), &(cl[idx+1]), (sz - idx)*sizeof(uint));
    }
    setClSize(cl, sz);
    cl[sz] = 0;
} 

// this function is supposed to work on sorted clauses
inline void delLit(cla_t *cl, const uint lit) {
	int i;
    assert(cl != NULL);
    assert(lit > 0);
    i = findLit(cl, lit);
    assert(i > 0);
    delLitAt(cl, (uint)i);
}

inline void replaceLit(cla_t *cl, const uint l1, const uint l2) {
	int i;
    assert(cl != NULL);
    assert(l1 > 0 && l2 > 0);
    i = findLit(cl, l1);
    assert(i > 0);
    assert(!isEquiv(cl) || cl[i] != cl[getClSize(cl)]);
    cl[i] = l2;
}

#ifdef DEBUG
void printCNFClause(char *msg, cla_t *c, FILE *fp) {
    uint i, csz = getClSize(c);
    uint to_free = (msg == NULL);
    if (to_free) {
        msg = xmalloc(sizeof(256));
    }
    if (isEquiv(c)) {
        writeStream("c ", fp);
    }
    if (csz == 0) writeStream("c ELIMINATED CLAUSE ", fp);
    for(i = 1; i < csz; i++) {
        uint lit = c[i];
        if (!isEquiv(c)) {
            sprintf(msg, "%d ", getPhase(lit) ? (int)getVar(lit) : 
                    -((int)getVar(lit)));
        } else {
            sprintf(msg, "%u ", lit);
        }
        writeStream(msg, fp);
    }
    if (isEquiv(c)) {
        uint v = c[csz];
        if (v) {
            if (isTautology(c)) {
                sprintf(msg, "[sgn = %u, %u - T] ", getEquivSgn(c), v);
            } else {
                sprintf(msg, "[sgn = %u, %u - D] ", getEquivSgn(c), v);
            }
        } else {
            sprintf(msg, "[sgn = %u] ", getEquivSgn(c));
        }
        writeStream(msg, fp);
    }
    writeStream("0\n", fp);
    if (to_free) {
        free(msg);
    }
}
#endif

// computes a dif b for equivalency clauses, for example:
// a = {1, 2, 3, 5}, b = {2, 3, 7}
// a dif b = {1, 5, 7}
// nlt = {7}    (new literals)
// ins = {2, 3} (intersection)
inline void clDif(cla_t *a, cla_t *b, uint *dif, uint *nlt, uint *ins) {
    uint sa = getClSize(a);
    uint sb = getClSize(b);
    uint i, j, di = 1, ni = 1, ii = 1;
    assert(sa > 1 && sb > 1);
    assert(checkClSort(a));
    assert(checkClSort(b));
    for (i = 1, j = 1; i < sa && j < sb;) {
        if (a[i] < b[j]) {
            dif[di++] = a[i];
            nlt[ni++] = a[i++];
        } else if (b[j] < a[i]) {
            dif[di++] = b[j++];
        } else {
            ins[ii++] = a[i++];
            j++;
        }
    } 
    for (; i < sa; i++) {
        dif[di++] = a[i];
        nlt[ni++] = a[i];
    }
    for (; j < sb; j++) {
        dif[di++] = b[j];
    }
    dif[0] = di;
    nlt[0] = ni;
    ins[0] = ii;
}
 
inline uint isUnitSz(const uint sz) {
    return sz == 2;
}

inline uint isBinSz(const uint sz) {
    return sz == 3;
}

inline uint isUnitCl(const cla_t *cl) {
    assert(cl != NULL);
    return isUnitSz(getClSize(cl));
}

inline uint isBinCl(const cla_t *cl) {
    assert(cl != NULL);
    return isBinSz(getClSize(cl));
}

/*
inline uint isTerCl(const cla_t *cl) {
    assert(cl != NULL);
    return getClSize(cl) == 4;
}// */

// create eqv clause from array of the form
// size [x y c] , where size = 4 in this case
// the created clause is of the form:
// size << 16 [x y c] dep_var, size = 4 in this case,
// x, y, and c are vars
// dep_var is initially zero
cla_t *formEqvCla(uint *a, uint sgn) {
    uint size = a[0];
    cla_t *n = xmalloc((size + 1) * sizeof(uint));
    memcpy(n, a, size * sizeof(uint));
    n[0] = 0;
    n[size] = 0;
    setClSize(n, size);
    markEquiv(n);
    setEquivSgn(n, sgn);
    return n;
}

cla_t *getNewClause(uint *a, uint size) {
	cla_t *cl;
    assert(size - 1 <= MAX_CL_SIZE);
    cl = xmalloc((size + 1) * sizeof(uint));
    cl[0] = 0;
    setClSize(cl, size + 1);
    assert(!isLearned(cl));
    assert(getEquivSgn(cl) == 0);
    memcpy(&(cl[1]), a, size*sizeof(uint));
    assert(getClSize(cl) > 1);
    return cl;
}

// size << 16 [x y c] dep_var, size = 4 in this case,
// x, y, and c are vars
// dep_var is initially zero
cla_t *getNewEqvCla(cla_t *c, uint sgn) {
    uint i, cs = getClSize(c);
    cla_t *n = xmalloc((cs + 1) * sizeof(uint));
    assert(!isEquiv(c));
    assert(!isLearned(c));
    assert(!isTautology(c));
    memcpy(n, c, cs * sizeof(uint));
    n[cs] = 0;
    markEquiv(n);
    setEquivSgn(n, sgn);
    assert(getEquivSgn(n) == sgn);
    assert(isEquiv(n));
    setClSize(n, cs);
    assert(getEquivSgn(n) == sgn);
    assert(isEquiv(n));
    for (i = 1; i < cs; i++) {
        n[i] >>= 1;
    }
    return n;
}        
#ifdef DEBUG
char checkClSort(cla_t *c) {
    uint i, x = 0, csz = getClSize(c);
    for (i = 1; i < csz; i++) {
        if (i > 1) {
            assert(x < c[i]);
        }
        x = c[i];
    }
    return TRUE;
}
#endif

inline uint getVar(const uint lit) {
    assert (lit);
    return lit >> 1;
}

inline uint getPhase(const uint lit) {
    assert(lit);
    return lit & 1;
}

#ifdef DEBUG
void arrayPrn(uint *x, uint size) {
    uint i;
    printf("A[");
    for (i = 0; i < size; i++) {
        int y;
        y = (x[i] == 0) ? 0 :
            getPhase(x[i]) ? (int)getVar(x[i]) : -(int)getVar(x[i]);
        printf("%d ", y);
    }
    printf("]\n");
}
#endif

void arrayClPrn(cla_t *cl) {
	uint i;
    uint sz = getClSize(cl);
    printf("A[");
    for (i = 0; i < sz; i++) {
        if (i == 0) {
            printf("%u ", cl[i]);
        } else {
            int y = getPhase(cl[i]) ? (int)getVar(cl[i]) : 
                -((int)getVar(cl[i]));
            printf("%d ", y);
        }
    }
    printf("]\n");
}

#ifdef COLLECT_STATS
stats_t *computeStats(stats_t *s) {
    s->hcAvgBranches = (s->hcNumHcubes) ?
        (float)((double)s->hcTotBranches/s->hcNumHcubes) : 0.0f;
    s->hcAvgNodes = (s->hcNumHcubes) ?
        (float)((double)s->hcTotNodes/s->hcNumHcubes) : 0.0f;
    s->hcAvgUsedBranches = (s->hcNumHcubes) ?
        (float)((double)s->hcUseBranches/s->hcNumHcubes) : 0.0f;
    s->hcAvgUsedNodes = (s->hcNumHcubes) ? 
        (float)((double)s->hcUseNodes/s->hcNumHcubes) : 0.0f;
    s->hcFlipBranchDensity = (s->cnfNum) ?
        (float)((double)s->hcTotBranches/s->cnfNum) : 0.0f;
    s->hcFlipNodeDensity = (s->cnfNum) ?
        (float)((double)s->hcTotNodes/s->cnfNum) : 0.0f;
    s->impAvgChainLen = (s->impNumChains) ? 
        (float)((double)s->impNumImps/s->impNumChains) : 0.0f;
    s->impAvgSchBlockSz = (s->impBlocksScheduled) ? 
        (float)((double)s->impTotalBlockSz/s->impBlocksScheduled) : 0.0f;
    s->cnfUIPAvgSz = (s->cnfUIPs) ? 
        (float)((double)s->cnfUIPCumulativeSz/s->cnfUIPs) : 0.0f;
    s->cnfDECAvgSz = (s->cnfDECs) ? 
        (float)((double)s->cnfDECCumulativeSz/s->cnfDECs) : 0.0f;
    return s;
}
#endif

// ************* UNIQUE MARKING TABLE ********************************

enum unit_table_e {WRITE_SUCCESS = 127};

table_t *initTable(uint elements, uint cell_size) {
	uint cell;
	table_t *t;
    assert(cell_size <= 4 && cell_size != 0);
    cell = (cell_size > 2) ? (uint)4 : cell_size;
    t = xcalloc(1, sizeof(table_t) + elements * sizeof(UTABLE_FIELD));
    t->cell_size = cell;
    t->mask = (unsigned char)((1 << cell) - 1);
    t->cnt =  (UTABLE_FIELD)(1 << cell);
    t->size = elements;
    return t;
}

void freeTable(table_t *t) {
    free(t);
}

inline void incrTimeStamp(table_t *t) {
    t->cnt += (1 << t->cell_size);
    if (!(t->cnt & ~t->mask)) {
        memset(t->a, 0, t->size * sizeof(UTABLE_FIELD));
        t->cnt += (1 << t->cell_size);
    }
}

inline uint readTable(const table_t *t, const uint idx) {
    assert(idx < t->size && "Index out of bounds");
    return (uint)t->a[idx] & t->mask;
}

inline int checkTable(const table_t *t, const uint idx) {
    uint x;
    assert(idx < t->size);
    x = t->a[idx];
    return ((x ^ t->cnt) > t->mask) ? -1 : (int)(x & t->mask);
}

inline void resetTableElem(table_t *t, const uint idx) {
    assert(idx < t->size);
    t->a[idx] = (UTABLE_FIELD)0;
}

// unique table write function
inline uint writeTable(table_t *t, const uint idx, const uint value) {
    assert(!(value & ~t->mask) && "About to discard some bits!");
    assert(idx < t->size);
    // if there's already a value with the most recent stamp,
    // return it without writing any new data 
    if (t->a[idx] >= t->cnt) {
        return readTable(t, idx);
    }
    // otherwise write new data and return UCHAR_MAX;
    t->a[idx] = t->cnt | (UTABLE_FIELD)value;
    return WRITE_SUCCESS;
}


// ******************** FLEX ARRAY ********************************

// returns the current size of the array - any append to the array
// or array reset will make this value invalid
inline uint getFlexArraySize(const flexarray_t *f) {
    return f->idx;
}

inline uint *getFlexArray(const flexarray_t *f) {
    return f->a;
}

#ifdef DEBUG
char checkSortFlexArray(const flexarray_t *f) {
    uint *a = getFlexArray(f);
    uint x = 0, i, sz = getFlexArraySize(f);
    for (i = 0; i < sz; i++) {
        if (i == 0) {
            x = a[i];
        } else {
            assert(x <= a[i]);
            x = a[i];
        }
    }
    return TRUE;
}
#endif

inline void resetFlexArray(flexarray_t *f) {
    f->idx = 0;
}

/*
inline void shrinkFlexArray(flexarray_t *f, uint tresh) {
    assert(f->idx == 0);
    if (f->size >= tresh) {
        f->a = (uint *)xrealloc((char *)f->a, 
                FLEX_SHRINKED_SIZE * sizeof(uint));
        f->size = FLEX_SHRINKED_SIZE;
    }
}// */

void freeFlexArray(flexarray_t *f) {
    free (f->a);
    free(f);
}

void delFlexElem(flexarray_t *f, uint i) {
    assert(i < f->idx);
    f->idx--;
    swap((int *)f->a, (int)i, (int)f->idx);
}

void appendToFlexArray(flexarray_t *f, uint x) {
    if (f->idx >= f->size) {
        f->size <<= 1;
        f->a = (uint *)xrealloc((char *)f->a, f->size * sizeof(uint));
    }
    f->a[(f->idx)++] = x;
}

flexarray_t *createFlexArray(uint *a, uint sz) {
    flexarray_t *f = xmalloc(sizeof(flexarray_t));
    f->a = xmalloc(sz * sizeof(uint));
    memcpy(f->a, a, sz * sizeof(uint));
    f->size = f->idx = sz;
    return f;
}

void qsortFlexArray(flexarray_t *f) {
    cmp_qsort(f->a, f->idx, 4, ucmp);
    assert(checkSortFlexArray(f));
}

inline int searchFlexArray(flexarray_t *f, uint x) {
    assert(checkSortFlexArray(f));
    return binsearch((int *)f->a, (int)f->idx, (int)x);
}

/*
inline int exorsearchFlex(flexarray_t *f, uint x) {
    return exorsearch((int *)f->a, (int)f->idx, (int)x);
}// */

int bfsearchFlexArray(flexarray_t *f, uint x, uint inc) {
    uint i;
    for (i = 0; i < f->idx; i+=inc) {
        if (f->a[i] == x) {
            return (int)i;
        }
    }
    return -1;
}

/*
void reverseFlexArray(flexarray_t *f) {
    uint *a = getFlexArray(f);
    uint i = 0, t = getFlexArraySize(f) - 1;
    while (t > i) {
        swap((int *)a, (int)i, (int)t);
        i++;
        t--;
    }
}// */

inline flexarray_t *initFlexArray(uint size) {
    flexarray_t *f = xmalloc(sizeof(flexarray_t));
    f->size = size;
    f->idx = 0;
    f->a = xcalloc(1, size * sizeof(uint));
    return f;
}

inline flexarray_t *partialInitFlex(void) {
    flexarray_t *f = xmalloc(sizeof(flexarray_t));
    f->a = NULL;
    return f;
}

inline void insertInFlex(flexarray_t *f, uint *a, uint sz) {
    f->size = sz;
    f->idx = sz;
    f->a = a;
}

inline void partialFreeFlex(flexarray_t *f) {
    free(f);
}

/*
inline void swapFlexArray(flexarray_t *a, flexarray_t *b) {
    // swap arrays
    uint tm, *tmp;
    tmp = a->a;
    a->a = b->a;
    b->a = tmp;
    // swap sizes
    tm = a->size;
    a->size = b->size;
    b->size = tm;
    // swap indices
    tm = a->idx;
    a->idx = b->idx;
    b->idx = tm; 
}

// deep copy of flex array, returns a new flex array
flexarray_t *copyFlexArray(flexarray_t *f) {
    flexarray_t *c = initFlexArray(getFlexArraySize(f));
    c->idx = f->idx;
    memcpy(c->a, f->a, f->idx * sizeof(uint));
    return c;
}// */

void multiAppendFlex(flexarray_t *from, flexarray_t *to) {
    uint ns = from->idx + to->idx;
    if (ns >= to->size) {
        to->a = (uint *)xrealloc((char *)to->a, ns * sizeof(uint));
    }
    to->size = ns;
    memcpy(&(to->a[to->idx]), from->a, from->idx * sizeof(uint));
    to->idx += from->idx;
}

/*
void printFlex(flexarray_t *f) {
    uint *a = getFlexArray(f);
    uint i, sz = getFlexArraySize(f);
    for (i =0; i < sz; i++) {
        printf("%u ", a[i]);
    }
    printf("\n");
}

void printFlexLits(flexarray_t *f) {
    uint *a = getFlexArray(f);
    uint i, sz = getFlexArraySize(f);
    for (i = 0; i < sz; i++) {
        uint lit = a[i];
        uint v = a[i] >> 1;
        printf("%d ", (LSB(lit)) ? (int)v : -((int)v));
    }
    printf("0 \n");
}

// intersection of two flexarrays, resulting array
// is always properly sorted if inputs are sorted
void intersectFlexArrays(flexarray_t *fa, flexarray_t *fb, 
        flexarray_t *res) {
	uint *a, *b, i, j;
    assert(getFlexArraySize(res) == 0);
    assert(checkSortFlexArray(fa));
    assert(checkSortFlexArray(fb));
    a = getFlexArray(fa);
    b = getFlexArray(fb);
    for (i = 0, j = 0; i < fa->idx && j < fb->idx;) {
        if (a[i] < b[j]) {
            i++;
        } else if (b[j] < a[i]) {
            j++;
        } else { // a[i] == b[j]
            assert(a[i] != 0);
            appendToFlexArray(res, a[i]);
            i++; j++;
        }
    }
    assert(getFlexArraySize(res) <= (uint)_min((int)fa->size, (int)fb->size));
    assert(checkSortFlexArray(res));
}// */

// *********** BCP QUEUE - queue for unit literals *******************

bcpq_t *initBcpq(uint size) {
    bcpq_t *m = xmalloc(sizeof(bcpq_t) + 2 * size * sizeof(uint));
    m->reasons = (cla_t **)xmalloc((size + 1) * sizeof(cla_t *));
    m->ridx = 0;
    m->widx = 0;
    m->t = initTable(size + 1, 2);
    return m;
}

inline int bcpqEmpty(bcpq_t *m) {
    return m->widx == m->ridx;
}

void freeBcpq(bcpq_t *m) {
    freeTable(m->t);
    free(m->reasons);
    free(m);
}

void bcpqReset(bcpq_t *m) {
    m->widx = 0;
    m->ridx = 0;
    incrTimeStamp(m->t);
}

// *************** ASSIGNEMENT QUEUE *******************************

assign_t *initAQ(uint size) {
    assign_t *aq = xmalloc(sizeof(assign_t) + size * sizeof(uint));
    aq->current_dlev = 0;
    aq->todo = xcalloc(1, 2 * size * sizeof(uint));
    aq->dec = xmalloc((size + 1) * sizeof(uint));
    aq->tptr = 0;
    aq->trd_ptr = 0;
    aq->size = size << 1;
    aq->didx = 0;
    aq->aidx = 0;
    return aq;
}

inline void appendAQ(assign_t *aq, uint lit, uint dlev, char dec) {
    assert(lit);
    assert(dlev > 0);
    aq->a[aq->aidx++] = lit;
    aq->current_dlev = dlev;
    if (dec && dlev > 0) {
        aq->dec[++aq->didx] = lit;
    }
}

inline void resetAQ(assign_t *aq) {
    aq->current_dlev = 0;
    aq->didx = 0;
    aq->aidx = 0;
}

inline void scheduleAQ(assign_t *aq, uint lit, uint ext) {
    assert(aq->tptr < (aq->size - 1));
    aq->todo[aq->tptr++] = lit;
    aq->todo[aq->tptr++] = ext;
}
 
inline uint getDecVar(assign_t *aq, uint dlev) {
    assert(dlev > 0);
    assert(dlev <= aq->current_dlev);
    return getVar(aq->dec[dlev]);
}

flexarray_t *getAQ(assign_t *aq) {
    return createFlexArray(aq->a, aq->aidx);
}

inline void resetScheduledAQ(assign_t *aq) {
    aq->tptr = 0;
    aq->trd_ptr = 0;
}

inline uint numScheduledAQ(assign_t *aq) {
    return (aq->tptr - aq->trd_ptr) >> 1;
}

void freeAQ(assign_t *aq) {
    free(aq->todo);
    free(aq->dec);
    free(aq);
}

// returns type of scheduled literal
inline uint getScheduledAQ(assign_t *aq, uint *lit, uint *ext) {
	int cmp;
	uint invcmp, lt, xt;
    assert(aq->trd_ptr <= aq->tptr);
    cmp = (aq->tptr == aq->trd_ptr);
    invcmp = ~((uint)-cmp);
    lt = aq->todo[aq->trd_ptr] & invcmp;
    assert(!cmp || lt == 0);
    aq->trd_ptr += (!cmp);
    xt = aq->todo[aq->trd_ptr] & invcmp;
    assert(!cmp || xt == 0);
    aq->trd_ptr += (!cmp);
    *lit = lt;
    *ext = xt;
    return (uint)_umin((uint)xt, 2);
}

inline uint numAssignedAQ(assign_t *aq) {
    return aq->aidx;
}
            
inline uint declev(assign_t *aq) {
    return aq->current_dlev;
}

uint popAQ(assign_t *aq) {
    if (aq->aidx > 0) {
        uint lit = aq->a[--aq->aidx];        
        aq->didx -= (lit == aq->dec[aq->didx]);
        return lit;
    } else return 0;
}

#endif //H_DATAS

/*
 * Revision history:
 *
 * $Log: datas.h,v $
 * Revision 1.11  2005/03/29 05:29:02  root
 * Fixed a bug in getScheduledAQ
 *
 * Revision 1.10  2005/02/21 22:32:30  root
 * SAT competition release.
 *
 * Revision 1.9  2005/02/09 01:59:25  root
 * Minor modification.
 *
 * Revision 1.8  2004/12/23 23:07:34  root
 * Optimized B-cubing implementation (now the algorithm picks the
 * variable with the highest priority as the new branch node and
 * search backtracks over unused BCT branch nodes). Fixed some
 * minor bugs. Code passes splint -weak check.
 *
 * Revision 1.6  2004/12/13 07:44:51  root
 * Adapted the code to VC C++ .NET compiler and 
 * made some minor optimizations.
 *
 * Revision 1.5  2004/12/11 04:46:11  root
 * Enofrced stricter c89 standard conformance.
 *
 * Revision 1.4  2004/12/09 18:46:17  root
 * Moved some functions from other files to datas.h
 *
 * Revision 1.3  2004/10/31 19:08:26  root
 * Minor modifications of unique table data structure.
 *
 * Revision 1.2  2004/10/24 00:04:22  root
 * Reduced the greediness of heuristic function.
 *
 * Revision 1.1  2004/10/12 23:31:15  root
 * Initial import.
 *
 *
 */
