/* Author: Shane Saunders
 * http://nz.cosc.canterbury.ac.nz/tad.takaoka/alg/heaps/heaps.html
 *
 * Modified by: Domagoj Babic 2005.
 *
 * $Id: fheap.h,v 1.6 2005/02/21 23:44:09 root Exp $
 *
 */

/***********************************************************************
*
*  The author of most of this code is Shane Saunders. Please see his 
*  webpage for licensing terms and copyright notice. 
*
*  THE SOFTWARE IS PROVIDED "AS-IS" AND Domagoj Babic MAKES NO 
*  REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED, BY WAY OF 
*  EXAMPLE, BUT NOT LIMITATION. Domagoj Babic MAKES NO 
*  REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS 
*  FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF THE SOFTWARE 
*  WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS TRADEMARKS OR OTHER 
*  RIGHTS. Domagoj Babic SHALL NOT BE LIABLE FOR ANY LIABILITY OR 
*  DAMAGES WITH RESPECT TO ANY CLAIM BY LICENSEE OR ANY THIRD PARTY 
*  ON ACCOUNT OF, OR ARISING FROM THE LICENSE, OR ANY SUBLICENSE OR 
*  USE OF THE SOFTWARE OR ANY SERVICE OR SUPPORT.
*
*  Domagoj Babic shall not be liable under any circumstances for any 
*  direct, indirect, special, incidental, or consequential damages with 
*  respect to any claim by USER or any third party on account of or 
*  arising from the use, or inability to use, this software or its 
*  associated documentation, even if the Author has been advised of the 
*  possibility of those damages.
*
*  The above copyright notice must appear in all copies of the software.
*************************************************************************
*
*  Support notice: Unfortunatelly, I can't guarantee any support. The 
*  solver has been written exclusively for research and experimental 
*  purposes, so documentation and comments are scarce. Most likely, 
*  this will not change in the near future, but feel free to check my 
*  webpage occasionally, as new releases (if there will be any) will 
*  be made available at my website:
*  http:/www.cs.ubc.ca/~babic/
*
***********************************************************************/

#ifndef FHEAP_H
#define FHEAP_H

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include "fheap.h"
#include "utils.h"

/*** Definitions of structure types. ***/

/* The structure type for Fibonacci heap nodes.
 *
 * Nodes have the following pointers:
 * parent      - a pointer to the nodes parent node (if any).
 * child       - a pointer to a child node (typically the highest 
 *               rank child).
 * left, right - sibling pointers which provide a circular doubly 
 *               linked list containing all the parents nodes 
 *               children.
 * The remaining structure fields are:
 * rank        - the nodes rank, that is, the number of children it 
 *               has.
 * `key'       - the nodes key.
 * vertex_no   - the number of the graph vertex that the node 
 *               corresponds to. Vertex numbering in the graph 
 *               should be: 1, 2, 3, ... max_vertex.
 */
typedef struct fheap_node {
    struct fheap_node *parent;
    struct fheap_node *left, *right;
    struct fheap_node *child;
    int rank;
    int marked;
    long key;
    int vertex_no;
} fheap_node_t;

/* The structure type for a Fibonacci heap.
 *
 * trees - An array of pointers to trees at root level in the heap.  
 *         Entry i in the array points to the root node of a tree 
 *         that has nodes of dimension i on the main trunk.
 * nodes - An array of pointers to nodes in the heap.  Nodes are 
 *         indexed according to their vertex number.  This array can 
 *         then be used to look up the node for corresponding to a 
 *         vertex number, and is useful when freeing space taken up 
 *         by the heap.
 * max_nodes - The maximum number of nodes allowed in the heap.
 * max_trees - The maximum number of trees allowed in the heap 
 *             (calculated from max_nodes).
 * n     - The current number of nodes in the heap.
 * value - The binary value represented by trees in the heap.
 *         By maintaining this it is easy to keep track of the maximum 
 *         rank tree in the heap.
 */
typedef struct fheap {
    fheap_node_t **trees;
    fheap_node_t **nodes;
    int max_nodes, max_trees, n, value;
    long key_comps;
} fheap_t;

/*** Function prototypes. ***/

/* Fibonacci heap functions. */
/* Note that fheap_t pointers are used so that function definitions are 
 * compatible with those of other heaps.  This allows any type heap to 
 * be given as an argument to a particular algorithm.  It is up to the 
 * user to ensure the correct heap type is passed to the given 
 * functions. */

void fh_dump_nodes(fheap_node_t *ptr, int level);
void fh_meld(fheap_t *h, fheap_node_t *tree_list);

// change the node's key and insert it into the heap 
void fh_node_insert(fheap_t *h, fheap_node_t *n, long k);

// create and initialize a new heap node
fheap_node_t *fh_create(int vertex_no);

/* fh_alloc() - creates and and returns a pointer to a F-heap which can 
 * contain up to max_nodes nodes. */
fheap_t *fh_alloc(int max_nodes);

/* fh_free() - destroys the heap pointed to by h, freeing up any space 
 * that was used by it. */
void fh_free(fheap_t *h);

/* fh_insert() - creates and inserts new a node representing vertex_no 
 * with key k into the heap h. 
void fh_insert(fheap_t *h, int vertex_no, long k);
// */

/* fh_delete_min() - deletes the minimum node from the heap pointed to 
 * by h and returns its vertex number. */
int fh_delete_min(fheap_t *h);

/* fh_decrease_key() - decreases the key used for vertex, vertex_no, to
 * new_value.  No check is made to ensure that new_value is in-fact less 
 * than the current value so it is up to the user of this function to 
 * ensure that it is. */
void fh_decrease_key(fheap_t *h, int vertex_no, long new_value);

long fh_max_priority(fheap_t *h);

// fh_alloc() - creates and and returns a pointer to a F-heap which 
// can contian up to max_nodes nodes.
fheap_t *fh_alloc(int max_nodes) {
    fheap_t *h;
    /* Create the heap. */
    h = xmalloc(sizeof(fheap_t));
    h->max_trees = (int)(1.0+1.44*log((double)max_nodes)/log(2.0));
    h->max_nodes = max_nodes;
    h->trees = xcalloc((size_t)h->max_trees, sizeof(fheap_node_t *));
    h->nodes = xcalloc((size_t)max_nodes, sizeof(fheap_node_t *));
    h->n = 0;
    /* The value of the heap helps to keep track of the maximum rank 
     * while nodes are inserted or deleted. */
    h->value = 0;
    /* For experimental purposes, we keep a count of the number of key
     * comparisons.  */
    h->key_comps = 0;
    return h; 
}

// fh_free() - destroys the heap pointed to by h, freeing up any space 
// that was used by it.
void fh_free(fheap_t *h) {
    free(h->nodes);
    free(h->trees);
    free(h);
}

// creates a new heap node 
fheap_node_t *fh_create(int vertex_no) {
    fheap_node_t *ptr = (fheap_node_t *)xmalloc(sizeof(fheap_node_t));
    ptr->child = NULL;
    ptr->left = ptr->right = ptr;
    ptr->rank = 0;
    ptr->vertex_no = vertex_no;
    return ptr;
}

/* fh_insert() - creates and inserts new a node representing vertex_no 
// with key k into the heap h.
void fh_insert(fheap_t *h, int vertex_no, long k) {
    fheap_node_t *ptr;
    // Create an initialise the new node.
    ptr = (fheap_node_t *)xmalloc(sizeof(fheap_node_t));
    ptr->child = NULL;
    ptr->left = ptr->right = ptr;
    ptr->rank = 0;
    ptr->vertex_no = vertex_no;
    ptr->key = k;
    // Maintain a pointer vertex_no's new node in the heap.
    h->nodes[vertex_no] = ptr;
    // Meld the new node into the heap.
    fh_meld(h, ptr);
    // Update the heaps node count. 
    h->n++;
}// */

// inserts already existing node, if it's already in the heap,
// the function will decrease the key only if it's smaller than
// the current one
void fh_node_insert(fheap_t *h, fheap_node_t *n, long k) {
    int vno = n->vertex_no;
    if (h->nodes[vno] == n) {
        // quietly returns...
        if (k < n->key) fh_decrease_key(h, n->vertex_no, k);
        return;
    }
    n->key = k;
    h->nodes[vno] = n;
    fh_meld(h, n);
    h->n++;
}

// fh_delete_min() - deletes the minimum node from the heap pointed 
// to by h and returns its vertex number.
int fh_delete_min(fheap_t *h) {
    fheap_node_t *min_node, *child, *next;
    long k, k2;
    int r, v, vertex_no;
    /* First we determine the maximum rank in the heap. */
    v = h->value;
    r = -1;
    while(v) {
        v = v >> 1;
        r++;
    };
    // if heap is empty, return 0
    if (r == -1) return 0;
    /* Now determine which root node is the minimum. */
    min_node = h->trees[r];
    k = min_node->key;
    while(r > 0) {
        r--;
        next = h->trees[r];
        if(next) {
            if((k2 = next->key) < k) {
                k = k2;
                min_node = next;
            }
            h->key_comps++;
        }
    }
    // We remove the minimum node from the heap but keep a pointer 
    // to it.
    r = min_node->rank;
    h->trees[r] = NULL;
    h->value -= (1 << r);
    child = min_node->child;
    if (child) {
        fh_meld(h, child);
        min_node->child = NULL;
    }
    // Record the vertex no of the old minimum node before deleting it.
    vertex_no = min_node->vertex_no;
    h->nodes[vertex_no] = NULL;
    min_node->left = min_node->right = min_node;
    min_node->rank = 0;
    h->n--;
    return vertex_no;
}

long fh_max_priority(fheap_t *h) {
    fheap_node_t *min_node, *next;
    long k, k2;
    int r, v;
    /* First we determine the maximum rank in the heap. */
    v = h->value;
    r = -1;
    while(v) {
        v = v >> 1;
        r++;
    };
    // if heap is empty, return 0
    if (r == -1) return 0;
    /* Now determine which root node is the minimum. */
    min_node = h->trees[r];
    k = min_node->key;
    while(r > 0) {
        r--;
        next = h->trees[r];
        if(next) {
            if((k2 = next->key) < k) {
                k = k2;
                min_node = next;
            }
            h->key_comps++;
        }
    }
    return k;
}

// fh_decrease_key() - decreases the key used for vertex, vertex_no, 
// to new_value.  No check is made to ensure that new_value is 
// in-fact less than the current value so it is up to the user of 
// this function to ensure that it is.
void fh_decrease_key(fheap_t *h, int vertex_no, long new_value) {
    fheap_node_t *cut_node, *parent, *new_roots, *r, *l;
    int prev_rank;
    /* Obtain a pointer to the decreased node and its parent then 
     * decrease the nodes key.  */
    cut_node = h->nodes[vertex_no];
    parent = cut_node->parent;
    cut_node->key = new_value;
    /* No reinsertion occurs if the node changed was a root. */
    if(!parent) {
        return;
    }
    /* Update the left and right pointers of cut_node and its two 
     * neighbouring nodes. */
    l = cut_node->left;
    r = cut_node->right;
    l->right = r;
    r->left = l;
    cut_node->left = cut_node->right = cut_node;

    /* Initially the list of new roots contains only one node. */
    new_roots = cut_node;

    /* While there is a parent node that is marked a cascading cut 
     * occurs. */
    while(parent && parent->marked) {
        /* Decrease the rank of cut_node's parent an update its 
         * child pointer. */
        parent->rank--;
        if(parent->rank) {
            if(parent->child == cut_node) parent->child = r;
        }
        else {
            parent->child = NULL;
        }
        /* Update the cut_node and parent pointers to the parent. */
        cut_node = parent;
        parent = cut_node->parent;
        /* Update the left and right pointers of cut_nodes two 
         * neighbouring nodes. */
        l = cut_node->left;
        r = cut_node->right;
        l->right = r;
        r->left = l;
        /* Add cut_node to the list of nodes to be reinserted as 
         * new roots. */
        l = new_roots->left;
        new_roots->left = l->right = cut_node;
        cut_node->left = l;
        cut_node->right = new_roots;
        new_roots = cut_node;
    }

    /* If the root node is being relocated then update the trees[] 
     * array. Otherwise mark the parent of the last node cut. */
    if(!parent) {
        prev_rank = cut_node->rank + 1;
        h->trees[prev_rank] = NULL;
        h->value -= (1 << prev_rank);
    }
    else {
        /* Decrease the rank of cut_node's parent an update its 
         * child pointer. */
        parent->rank--;
        if(parent->rank) {
            if(parent->child == cut_node) parent->child = r;
        }
        else {
            parent->child = NULL;
        }
        parent->marked = 1;
    }

    /* Meld the new roots into the heap. */
    fh_meld(h, new_roots);
}



// Definitions of functions that are only visible within this file.

/* fh_meld() - melds  the linked list of trees pointed to by 
 * *tree_list into the heap pointed to by h. */
void fh_meld(fheap_t *h, fheap_node_t *tree_list) {
    fheap_node_t *first, *next, *node_ptr, *new_root, *temp, *temp2, *lc, *rc;
    int r;

    /* We meld each tree in the circularly linked list back into 
     * the root level of the heap.  Each node in the linked list 
     * is the root node of a tree. The circularly linked list 
     * uses the sibling pointers of nodes. This makes melding of 
     * the child nodes from a delete_min operation simple. */
    node_ptr = first = tree_list;
    do {
        /* Keep a pointer to the next node and remove sibling 
         * and parent links from the current node. node_ptr 
         * points to the current node. */
        next = node_ptr->right;
        node_ptr->right = node_ptr->left = node_ptr;
        node_ptr->parent = NULL;
        /* We merge the current node, node_ptr, by inserting it 
         * into the root level of the heap. */
        new_root = node_ptr;
        r = node_ptr->rank;
        /* This loop inserts the new root into the heap, possibly 
         * restructuring the heap to ensure that only one tree for 
         * each degree exists. */
        do {
            /* Check if there is already a tree of degree r in the 
             * heap. If there is then we need to link it with 
             * new_root so it will be reinserted into a new place 
             * in the heap. */
            if((temp = h->trees[r])) {
	            /* temp will be linked to new_root and relocated so 
                 * we no longer will have a tree of degree r. */
                h->trees[r] = NULL;
                h->value -= (1 << r);
	            /* Swap temp and new_root if necessary so that 
                 * new_root always points to the root node which 
                 * has the smaller key of the two. */
	            if(temp->key < new_root->key) {
                    temp2 = new_root;
                    new_root = temp;
                    temp = temp2;
                }
                h->key_comps++;
                /* Link temp with new_root, making sure that 
                 * sibling pointers get updated if rank is greater 
                 * than 0.  Also, increase r for the next pass 
                 * through the loop since the rank of new has
                 * increased. */
	            if(r++ > 0) {
                    rc = new_root->child;
                    lc = rc->left;
                    temp->left = lc;
                    temp->right = rc;
                    lc->right = rc->left = temp;
                }
                new_root->child = temp;
                new_root->rank = r;
                temp->parent = new_root;
                temp->marked = 0;
            } else {
            /* Otherwise if there is not a tree of degree r in the 
             * heap we allow new_root, which possibly carries moved 
             * trees in the heap, to be a tree of degree r in the 
             * heap. */
                h->trees[r] = new_root;
                h->value += (1 << r);;
                /* NOTE:  Because new_root is now a root we ensure 
                 * it is marked. */
                new_root->marked = 1;
	        }
            /* Note that temp will be NULL if and only if there was 
             * not a tree of degree r. */
        } while(temp);
        node_ptr = next;
    } while(node_ptr != first);
}

#endif // FHEAP_H

/*
 * Revision history:
 *
 * $Log: fheap.h,v $
 * Revision 1.6  2005/02/21 23:44:09  root
 * Minor update.
 *
 *
 * Revision 1.2  2004/12/09 18:47:43  root
 * Added fh_max_priority function.
 *
 * Revision 1.1  2004/10/12 23:31:15  root
 * Initial import.
 *
 *
 */


